/*
 * linux/arch/alpha/kernel/kgdb.c
 *
 * Copyright (c) 1995	David Mosberger (davidm@cs.arizona.edu)
 */
/*
 * This module provides low-level support for GDB and has been adapted
 * from sparc-stub.c.  Please refer to that file for more on the
 * history of the module.  The contributors seem to be: Hewlett-Packard,
 * Glenn Engel, and Cygnus Support (Stu Grossman).
 *
 * To enable debugger support, two things need to happen.  One, a call
 * to kgdb_set_debug_traps() is necessary in order to allow any
 * breakpoints or error conditions to be properly intercepted and
 * reported to gdb.  Two, a breakpoint needs to be generated to begin
 * communication.  This is most easily accomplished by a call to
 * kgdb_breakpoint().  kgdb_breakpoint() simulates a breakpoint by
 * executing a call_pal bpt.
 *
 *************
 *
 *    The following gdb commands are supported:
 *
 * command	    function				   Return value
 *
 *    g		    return the value of the CPU registers  hex data or ENN
 *    G		    set the value of the CPU registers	   OK or ENN
 *    Pn..=v..	    set CPU register n.. to value v..	   OK or ENN
 *
 *    mAA..AA,LLLL  Read LLLL bytes at address AA..AA	   hex data or ENN
 *    MAA..AA,LLLL: Write LLLL bytes at address AA.AA	   OK or ENN
 *
 *    c		    Resume at current address		   SNN	 (signal NN)
 *    cAA..AA	    Continue at address AA..AA		   SNN
 *
 *    s		    Step one instruction		   SNN
 *    sAA..AA	    Step one instruction from AA..AA	   SNN
 *
 *    k		    kill
 *
 *    ?		    What was the last sigval ?		   SNN	 (signal NN)
 *
 *   qXXXX	    Request info about XXXX.
 *   QXXXX=yyyy	    Set value of XXXX to yyyy.
 *   qOffsets	    Get section offsets.	     Text=xxx;Data=yyy;Bss=zzz
 *
 *		    Send text to stdout.  Only comes from  Otext
 *		    remote target.
 *
 * All commands and responses are sent with a packet which includes a
 * checksum.  A packet consists of
 *
 * $<packet info>#<checksum>.
 *
 * where
 * <packet info> :: <characters representing the command or response>
 * <checksum>    :: <two hex digits computed as modulo 256 sum of <packetinfo>>
 *
 * When a packet is received, it is first acknowledged with either '+' or '-'.
 * '+' indicates a successful transfer.  '-' indicates a failed transfer.
 *
 * Example:
 *
 * Host:                  Reply:
 * $m0,10#2a               +$00010203040506070809101112131415#42
 *
 * Responses can be run-length encoded to save space.  A '*' means
 * that the next characters is giving a repeat count in the range
 * 3..97.  A repeat count of 3 is encoded as a blank character (0x20)
 * and a repeat count of 97 is encoded as a tilde character (0x7e).
 * The repeat count stands for that many repititions of the character
 * preceding the '*'.  Note that this means that responses cannot
 * contain '*' (this might be a problem in the console-output
 * request).
 *
 * Example: "0* " means the same as "0000".
 */
#include <linux/config.h>

#include <linux/kernel.h>
#include <linux/ptrace.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/string.h>

#include <asm/system.h>
#include <asm/io.h>


#ifdef CONFIG_KGDB_TRACING

static struct trace_buf {
	u_long		id;
	u_long *	base;
	u_long *	volatile limit;
	u_long *	next;
	u_long *	flags;
} trace_buf = {0, 0, 0, 0};

#endif /* CONFIG_KGDB_TRACING */


extern void kgdb_enter (void);

struct ext_pt_regs {
	unsigned long	sp;
	unsigned long	r9;
	unsigned long	r10;
	unsigned long	r11;
	unsigned long	r12;
	unsigned long	r13;
	unsigned long	r14;
	unsigned long	r15;
	unsigned long	fpu[32];
	struct pt_regs	std;
};

/*
 * These defines must match the corresponding code in remote.c.
 */
#define RLE_MIN_COUNT	3
#define RLE_MAX_COUNT	(('~' - ' ') + RLE_MIN_COUNT)

#define BREAKINST	0x00000080	/* call_pal bpt */

/*
 * BUFMAX defines the maximum number of characters in inbound/outbound
 * buffers at least NUMREGBYTES*2 are needed for register packets.
 */
#define BUFMAX 2048

#define NUMREGS		(2*32+1)	/* # of registers */
#define NUMREGBYTES	(NUMREGS * 8)	/* # of bytes for registers */

enum regnames {
      V0,   T0,   T1,   T2,   T3,   T4,   T5,   T6,
      T7,   S0,   S1,   S2,   S3,   S4,   S5,   S6,
      A0,   A1,   A2,   A3,   A4,   A5,   T8,   T9,
     T10,  T11,   RA,  T12,   AT,   GP,   SP, ZERO,
      F0,   F1,   F2,   F3,   F4,   F5,   F6,   F7,
      F8,   F9,  F10,  F11,  F12,  F13,  F14,  F15,
     F16,  F17,  F18,  F19,  F20,  F21,  F22,  F23,
     F24,  F25,  F26,  F27,  F28,  F29,  F30,  F31,
      PC,  VFP
};

#define DBG_STACK		0x001
#define DBG_TRAP		0x002
#define DBG_IN_PACKETS		0x004
#define DBG_OUT_PACKETS		0x008
#define DBG_MEM			0x010
#define DBG_STORE_REG		0x020
#define DBG_CMDS		0x040
#define DBG_SINGLE_STEP		0x080
#define DBG_FUNCS		0x100
#define DBG_TRACING		0x200

#define DEBUG
#ifdef DEBUG
# define DBG(bit,args)	{if (kgdb_debug & (bit)) printk args;}
#else
# define DBG(bit,args)
#endif

int kgdb_debug	 = 0x000;		/* kgdb debugging level */

static const char hexchars[] = "0123456789abcdef";

static int		nsaved = 0;
static unsigned int	saved_insn[2];
static unsigned int	*saved_addr[2];

static int initialized = 0;	/* !0 means we've been initialized */
static char remcomInBuffer[BUFMAX];
static char remcomOutBuffer[BUFMAX];

/*
 * Serial I/O routines.  Define SERIAL_LINE either as 0 or 1 depending
 * on whether you want to use /dev/cua0 or /dev/cua1 for running kgdb.
 */
#define SERIAL_LINE	1
#define SERIAL_BASEADDR	(0x300 - SERIAL_LINE * 0x100)
#define SERIAL_PORT(p)	(SERIAL_BASEADDR + (p))


void putDebugChar(int ch)
{
	unsigned long flags;
	int line_status;

	save_flags(flags);
	cli();

	/* wait for transmitter holding register to empty: */
	do {
		line_status = inb(SERIAL_PORT(0xfd));
	} while((line_status & 0x60) != 0x60);

	/* send the character out: */
	outb(ch, SERIAL_PORT(0xf8));

	restore_flags(flags);
}


int getDebugChar(void)
{
	unsigned long flags;
	int line_status, ch;

	save_flags(flags);
	cli();

	/* wait for "data ready": */
	do {
		line_status = inb(SERIAL_PORT(0xfd));
	} while((line_status & 0x01) == 0);

	ch = inb(SERIAL_PORT(0xf8));

	restore_flags(flags);

	return ch;
}


/*
 * Look for Ctrl-C.  If we find one, trap into debugger.
 */
static void irq_handler(int irq, void * dev_id, struct pt_regs *regs)
{
	if (getDebugChar() == 0x03) {
		kgdb_enter();
	}
}


void initDebugIO(void)
{
	int comstat, hi, lo;

	/*
	 * Initialize the serial port.  JRP This is deliberately
	 * brain-dead hard-coded bad code so it'll be replaced!
	 */
	printk("kgdb: stealing /dev/ttyS%d and initializing it to 38400 baud",
	       SERIAL_LINE);

	comstat = inb(SERIAL_PORT(0xfb));	/* turn on dlab */
	outb(comstat | 0x80, SERIAL_PORT(0xfb));
	hi = inb(SERIAL_PORT(0xf9));		/* read current baudrate */
	lo = inb(SERIAL_PORT(0xf8));
	outb(comstat, SERIAL_PORT(0xfb));	/* turn off dlab */

	outb(0x83, SERIAL_PORT(0xfb));	/* no parity, 8 data, 1 stop bit */
	outb(0x00, SERIAL_PORT(0xf9));	/* 38400 baud */
	outb(0x03, SERIAL_PORT(0xf8));
	outb(0x03, SERIAL_PORT(0xfb));	/* turnoff divisor latch access bit */
	/*
	 * Prior to disabling interrupts, read the LSR and RBR
	 * registers of BOTH devices.
	 */
	comstat = inb(SERIAL_PORT(0xfd));	/* COM2 LSR */
	comstat = inb(SERIAL_PORT(0xf8));	/* COM2 RBR */
	outb(0x01, SERIAL_PORT(0xf9));	/* enable receive interrupts */

	comstat = inb(SERIAL_PORT(0xfb));
	comstat = inb(SERIAL_PORT(0xfd));
}


/*
 * Misc. support routines:
 */
static inline long kaddr_range(unsigned long addr, unsigned long size)
{
	return ((addr >> 41) == 0x7ffffe) && (((addr + size - 1) >> 41) == 0x7ffffe);
}


/*
 * Don't want to use libraries memset() to avoid infinite recursion:
 */
static void private_memset(void *addr, long value, int length)
{
	char *dst = addr;

	while (length--) *dst++ = value;
}


/*
 * Convert ch from a hex digit to an int:
 */
static int hex(unsigned char ch)
{
	if (ch >= 'a' && ch <= 'f') {
		return ch - 'a' + 10;
	} else if (ch >= '0' && ch <= '9') {
		return ch - '0';
	} else if (ch >= 'A' && ch <= 'F') {
		return ch - 'A' + 10;
	} else {
		return -1;
	}
}


/*
 * Scan for the sequence $<data>#<checksum>:
 */
static void getpacket(char *buffer)
{
	unsigned char checksum;
	unsigned char xmitcsum;
	int i;
	int count;
	unsigned char ch;

	do {
		/* wait around for the start character, ignore all others */
		while ((ch = getDebugChar()) != '$') {
			/* skip */
		}

		checksum = 0;
		xmitcsum = -1;

		count = 0;

		/* now, read until a # or end of buffer is found */
		while (count < BUFMAX) {
			ch = getDebugChar();
			if (ch == '#') {
				break;
			}
			checksum = checksum + ch;
			buffer[count] = ch;
			count = count + 1;
		}

		if (count >= BUFMAX) {
			continue;
		}

		buffer[count] = 0;

		if (ch == '#') {
			xmitcsum = hex(getDebugChar()) << 4;
			xmitcsum |= hex(getDebugChar());
			if (checksum != xmitcsum) {
				putDebugChar('-');	/* failed checksum */
			} else {
				putDebugChar('+');	 /* successful xfer */
				/*
				 * if a sequence char is present,
				 * reply the sequence ID:
				 */
				if (buffer[2] == ':') {
					putDebugChar(buffer[0]);
					putDebugChar(buffer[1]);
					/* remove sequence chars from buffer */
					count = strlen(buffer);
					for (i=3; i <= count; i++) {
						buffer[i-3] = buffer[i];
					}
				}
			}
		}
	} while (checksum != xmitcsum);

	DBG(DBG_IN_PACKETS, ("kgdb: got packet `$%s'\n", remcomInBuffer));
}


/*
 * Flush out whatever repeated characters we have accumulated
 * so far.
 */
static unsigned char *rle_flush(unsigned char *dst, unsigned char rch,
				int rcount)
{
	if (rcount >= RLE_MIN_COUNT) {
		/*
		 * We have a sequence that's long enough to warrant
		 * encoding, do so:
		 */
		*dst++ = '*';
		*dst++ = (rcount - RLE_MIN_COUNT) + ' ';
	} else {
		/*
		 * The repeated sequence is not long enough to warrant
		 * rle, so output the repeated characters directly:
		 */
		while (rcount--) {
			*dst++ = rch;
		}
	}
	return dst;
}


/*
 * Encode sequences of repeated characters by one copy of that
 * character, a star, and a repeat-count encoded as a ASCII character.
 */
static void
rle_encode(unsigned char *src)
{
	unsigned char *dst, ch, rch;
	int rcount;

	dst = src;
	rch = '\0';
	rcount = 0;
	while ((ch = *src++) != '\0') {
		if (ch == rch) {
			++rcount;
			if (rcount >= RLE_MAX_COUNT) {
				dst = rle_flush(dst, rch, RLE_MAX_COUNT);
				rcount -= RLE_MAX_COUNT;
			}
		} else {
			dst = rle_flush(dst, rch, rcount);
			*dst++ = ch;	/* first character is always output */
			rch = ch;
			rcount = 0;
		}
	}
	dst = rle_flush(dst, rch, rcount);
	*dst++ = '\0';
}


/*
 * Send the packet in buffer.
 */
static void
putpacket(unsigned char *buffer)
{
	unsigned char checksum;
	unsigned char ch;
	int count;

	DBG(DBG_OUT_PACKETS, ("kgdb: sending packet `%s'\n", buffer));

	rle_encode(buffer);
	DBG(DBG_OUT_PACKETS, ("kgdb: encoded packet `%s'\n", buffer));

	/*  $<packet info>#<checksum>. */
	do {
		putDebugChar('$');
		checksum = 0;
		count = 0;

		while ((ch = buffer[count]) != '\0') {
			putDebugChar(ch);
			checksum += ch;
			count += 1;
		} /* while */

		putDebugChar('#');
		putDebugChar(hexchars[checksum >> 4]);
		putDebugChar(hexchars[checksum & 0xf]);

		DBG(DBG_OUT_PACKETS, ("kgdb: checksum %#x\n", checksum));
	} while (getDebugChar() == '-');
}


/*
 * Convert the memory pointed to by mem into hex, placing result in
 * buf.  Return a pointer to the last char put in buf (null), in case
 * of mem fault, return 0.  If MAY_FAULT is non-zero, then we will
 * handle memory faults by returning a 0, else treat a fault like any
 * other fault in the stub.
 */
static unsigned char *mem2hex(unsigned char *mem, unsigned char *buf,
			      int count, int may_fault)
{
	unsigned long l, *lp, flags;
	unsigned char ch;
	int byte;

	DBG(DBG_FUNCS, ("kgdb: mem2hex(mem=%p,buf=%p,count=%d,may_fault=%d)\n",
			mem, buf, count, may_fault));

	lp = (unsigned long*) ((unsigned long) mem & ~0x7);
	byte = mem - (unsigned char*) lp;
	save_flags(flags);
	cli();
	l = *lp++;
	l >>= 8 * byte;
	while (count-- > 0) {
		draina();
		if (rdmces() & 0x7) {
			if (!may_fault) {
				printk("addr=%p\n", lp - 1);
				panic("kgdb: mem2hex faulted!");
			}
			wrmces(0x7);			/* reset machine check */
			restore_flags(flags);
			return 0;			/* return error */
		}
		ch = l & 0xff;
		*buf++ = hexchars[ch >> 4];
		*buf++ = hexchars[ch & 0xf];
		l >>= 8;
		if (++byte >= 8) {
			byte = 0;
			l = *lp++;
		}
	}
	*buf = 0;
	restore_flags(flags);
	return buf;
}


/*
 * Convert the hex array pointed to by buf into binary to be placed in
 * mem return a pointer to the character AFTER the last byte written
 */
static char *hex2mem(unsigned char *buf, unsigned char *mem, int count,
		     int may_fault)
{
	unsigned long l, *lp, flags;
	unsigned char ch;
	int byte;
	int i;

	lp = (unsigned long*) ((unsigned long) mem & ~0x7);
	byte = mem - (unsigned char*) lp;
	save_flags(flags);
	cli();
	l = *lp;
	for (i = 0; i < count; ++i) {
		draina();
		if (rdmces() & 0x7) {
			if (!may_fault) {
				panic("kgdb: hex2mem faulted!");
			}
			wrmces(0x7);			/* reset machine check */
			restore_flags(flags);
			return 0;			/* return error */
		}
		ch = hex(*buf++) << 4;
		ch |= hex(*buf++);
		l = (l & ~(0xffL << 8*byte)) | ((long) ch << 8*byte);
		if (++byte >= 8) {
			byte = 0;
			*lp++ = l;
			draina();
			if (rdmces() & 0x7) {
				wrmces(0x7);			/* reset machine check */
				restore_flags(flags);
				return 0;			/* return error */
			}
			l = *lp;
		}
	}
	if (byte) {
		*lp = l;
	}
	restore_flags(flags);
	return buf;
}


/*
 * While we find nice hex chars, build an int.
 * Return number of chars processed.
 */
static int hex2int(char **ptr, int *int_value)
{
	int num_chars = 0;
	int hex_value;

	*int_value = 0;

	while (**ptr) {
		hex_value = hex(**ptr);
		if (hex_value < 0) {
			break;
		}

		*int_value = (*int_value << 4) | hex_value;
		num_chars ++;

		(*ptr)++;
	}

	return num_chars;
}


/*
 * While we find nice hex chars, build a long.
 * Return number of chars processed.
 */
static int hex2long(char **ptr, long *long_value)
{
	int num_chars = 0;
	long hex_value;

	*long_value = 0;

	while (**ptr) {
		hex_value = hex(**ptr);
		if (hex_value < 0) {
			break;
		}

		*long_value = (*long_value << 4) | hex_value;
		num_chars ++;

		(*ptr)++;
	}

	return num_chars;
}


/*
 * Convert a register number to the corresponding pointer into
 * the saved state.  Returns 0 in the case of a bad register.
 */
static unsigned long *regno2addr(unsigned int regno, struct ext_pt_regs *regs)
{
	static unsigned long zero;
	unsigned long *addr;

	switch (regno) {
	      case V0...T7: addr = &regs->std.r0 + regno; break;
	      case S0...S6: addr = &regs->r9 + regno - S0; break;
	      case A3...AT: addr = &regs->std.r19 + regno - A3; break;
	      case A0...A2: addr = &regs->std.r16 + regno - A0; break;
	      case SP:	    addr = &regs->sp; break;
	      case PC:	    addr = &regs->std.pc; break;
	      case GP:	    addr = &regs->std.gp; break;
	      case ZERO:    addr = &zero; zero = 0L; break;
	      case F0...F31: addr = &regs->fpu[regno - F0];
	      default:	    addr = 0; break;
	}
	return addr;
}


static int set_breakpoint(struct ext_pt_regs *regs)
{
	int displ, reg_b, i;
	unsigned long *reg_b_addr, flags;
	unsigned int insn, op_code;

	if (!kaddr_range(regs->std.pc, sizeof(unsigned int))) {
	    return 0;
	}

	save_flags(flags);
	cli();
	insn = *(unsigned int*)regs->std.pc;
	draina();
	if (rdmces() & 0x7) {
		wrmces(0x7);			/* reset machine check */
		restore_flags(flags);
		return 0;			/* return error */
	}

	op_code = insn >> 26;
	DBG(DBG_SINGLE_STEP, ("kgdb: single-step over "));
	if (op_code >= 0x30) {
		/*
		 * It's a branch: instead of trying to figure out
		 * whether the branch will be taken or not, we'll put
		 * a breakpoint at either location.  This is simpler,
		 * more reliable, and probably not a whole lot slower
		 * than the alternative approach of emulating the
		 * branch (emulation can be tricky for fp branches).
		 */
		DBG(DBG_SINGLE_STEP, ("branch"));
		displ = ((int)(insn<<11)) >> 11;
		saved_addr[nsaved++] = (unsigned int*)regs->std.pc + 1;
		if (displ)
		  saved_addr[nsaved++] = (unsigned int*)regs->std.pc + 1 + displ;
	} else if (op_code == 0x1a) {
		/* it's a jump */
		DBG(DBG_SINGLE_STEP, ("jump"));
		reg_b = (insn >> 16) & 0x1f;
		reg_b_addr = regno2addr(reg_b, regs);
		if (reg_b_addr) {
			saved_addr[nsaved++] = (unsigned int*) *reg_b_addr;
		} else {
			restore_flags(flags);
			return 0;
		}
	} else {
		DBG(DBG_SINGLE_STEP, ("normal"));
		saved_addr[nsaved++] = (unsigned int*)regs->std.pc + 1;
	}

	for (i = 0; i < nsaved; ++i) {
		if (!kaddr_range((unsigned long)saved_addr[i], sizeof(unsigned int))) {
			nsaved = 0;
			restore_flags(flags);
			return 0;
		}
		saved_insn[i] = *saved_addr[i];
		draina();
		if (rdmces() & 0x7) {
			wrmces(0x7);			/* reset machine check */
			nsaved = 0;
			restore_flags(flags);
			return 0;			/* return error */
		}
		DBG(DBG_SINGLE_STEP, (" -> next_pc=%p", saved_addr[i]));
		*saved_addr[i] = BREAKINST;
		draina();
		if ((rdmces() & 0x7) || *saved_addr[i] != BREAKINST) {
			nsaved = 0;
			wrmces(0x7);			/* reset machine check */
			restore_flags(flags);
			return 0;			/* return error */
		}
	}
	DBG(DBG_SINGLE_STEP, ("\n"));
	restore_flags(flags);
	return 1;
}


/*
 * This function does all command processing for interfacing to gdb.  It
 * returns the new stack pointer.
 */
void kgdb_handle_exception(struct ext_pt_regs *regs)
{
	unsigned long addr, *reg_addr;
	int sigval, length, i;
	unsigned int regno;
	char *ptr;
#ifdef CONFIG_KGDB_TRACING
	static int break_pending = 0;
#endif

	/*
	 * Restore original instructions as quickly as possible---in
	 * the case the stub executes the same code.
	 */
	for (i = 0; i < nsaved; ++i) {
		/* restore original instruction */
		DBG(DBG_SINGLE_STEP,
		    ("kgdb: single-step restoring instruction %x "
		     "at %p\n", saved_insn[i], saved_addr[i]));
		*saved_addr[i] = saved_insn[i];
	}
	nsaved = 0;

	if (((u32*)regs->std.pc)[-1] != BREAKINST) {
		/* make pc point to restored insn: */
		regs->std.pc -= 4;
	}

	DBG(DBG_TRAP,
	    ("kgdb: handle_exception(regs=%p, sp=%lx, pc=%lx)\n",
	     regs, regs->sp, regs->std.pc));

#ifdef CONFIG_KGDB_TRACING
	if (trace_buf.limit && trace_buf.next < trace_buf.limit) {
		static u32 *prev_pc = 0;
		unsigned int insn, ra, rb, rc;
		unsigned long *next;
		long dist;

		next = trace_buf.next;

		if (prev_pc) {
			insn = *prev_pc;

			/*
			 * When tracing, st?_c would always fail but because interrupts
			 * are disabled and we're executing on a uniprocessor, we *know*
			 * that all st?_c execute successfully, so we just fake the
			 * right result here:
			 */
			switch (insn >> 26) {
			      case 0x2e: /* stl_c */
			      case 0x2f: /* stq_c */
				ra = (insn >> 21) & 0x1f;
				*regno2addr(ra, regs) = 1;
			}
		}

		insn = *(u32*)regs->std.pc;	/* get insn we are going to single-step over */
		/*
		 * Encode relative insn to execute next.  If new PC is farther
		 * away than fits into a 32 bit integer, we record it's absolute
		 * address first, with lsb set to 0.  The relative address that
		 * contains the instruction in the top 32 bit has the lsb set.
		 */
		dist = regs->std.pc - (long) prev_pc;
		if (dist < -(1L<<31) || dist >= (1L<<31)) {
			*next++ = regs->std.pc;
			dist = 0;
		}
		*next++ = (((unsigned long) insn)<<32) | ((unsigned long) dist & 0xffffffff) | 1L;
		prev_pc = (u32*) regs->std.pc;

		/* record state that next insn uses (potentially): */
		rc = (insn >>  0) & 0x1f;
		rb = (insn >> 16) & 0x1f;
		ra = (insn >> 21) & 0x1f;
		*next++ = *regno2addr(ra, regs);
		*next++ = *regno2addr(rb, regs);
		*next++ = *regno2addr(rc, regs);

		/*
		 * Set new breakpoint if at least 5 longs are left in
		 * the trace buffer:
		 */
		break_pending = 0;
		trace_buf.next = next;
		if (next >= trace_buf.limit - 5) {
		    return;
		}
		break_pending = 1;
		set_breakpoint(regs);
		if (insn == PAL_swpipl) {
		    /* turn swpipl into NOP: */
		    *(u32*)regs->std.pc = 0x47ff041f;
		}
		/*
		 * When tracing, st?_c would always fail but because
		 * interrupts are disabled and we're executing on a
		 * uniprocessor, we *know* that all st?_c execute
		 * without interference, so we just replace the
		 * conditional store with an unconditional one.
		 * Notice that set_breakpoint() saved away the
		 * original opcode, so we can patch up the result in
		 * the code above when the instruction returns.
		 */
		switch (insn >> 26) {
		  case 0x2e: /* stl_c */
		  case 0x2f: /* stq_c */
		      *(u32*)regs->std.pc -= 0x2 << 26;
		      break;
		}
		/* ensure I-stream is coherent with D-stream: */
		imb();
		return;
	} else if (break_pending) {
		imb();			/* flush breakpoint cancellation */
		break_pending = 0;	/* simply ignore last tracing breakpoint */
		return;
	}
#endif /* CONFIG_KGDB_TRACING */

	/*
	 * Reply to host that an exception has occurred.  Include the
	 * minimal state GDB needs in order to do anything useful.
	 * The easiest way to find out which registers should be sent
	 * is to add a printk to function read_register() in
	 * gdb/findvar.c and do an "si" after connecting to the stub.
	 */
	sigval = SIGTRAP;
	ptr = remcomOutBuffer;
	*ptr++ = 'T';
	*ptr++ = hexchars[sigval >>  4];
	*ptr++ = hexchars[sigval & 0xf];

	*ptr++ = hexchars[S6 >>  4];	/* s6 also serves as frame pointer */
	*ptr++ = hexchars[S6 & 0xf];
	*ptr++ = ':';
	ptr    = mem2hex((char*)&regs->r15, ptr, 8, 0);
	*ptr++ = ';';

	*ptr++ = hexchars[SP >>  4];
	*ptr++ = hexchars[SP & 0xf];
	*ptr++ = ':';
	ptr    = mem2hex((char*)&regs->sp, ptr, 8, 0);
	*ptr++ = ';';

	*ptr++ = hexchars[PC >>  4];
	*ptr++ = hexchars[PC & 0xf];
	*ptr++ = ':';
	ptr    = mem2hex((char*)&regs->std.pc, ptr, 8, 0);
	*ptr++ = ';';

	/*
	 * GDB wants VFP --- even though it's a computed value:
	 */
	*ptr++ = hexchars[VFP >>  4];
	*ptr++ = hexchars[VFP & 0xf];
	*ptr++ = ':';
	private_memset(ptr, '0', 8 * 2); ptr += 8 * 2;
	*ptr++ = ';';

	*ptr++ = '\0';
	DBG(DBG_OUT_PACKETS, ("kgdb: sending trap info `%s'\n",
			      remcomOutBuffer));
	putpacket(remcomOutBuffer);

	while (1) {
		remcomOutBuffer[0] = '\0';

		getpacket(remcomInBuffer);
		DBG(DBG_CMDS, ("kgdb: got command `%c'\n", remcomInBuffer[0]));
		switch (remcomInBuffer[0]) {
		      case '?':	/* return last signal */
			remcomOutBuffer[0] = 'S';
			remcomOutBuffer[1] = hexchars[sigval >> 4];
			remcomOutBuffer[2] = hexchars[sigval & 0xf];
			remcomOutBuffer[3] = 0;
			break;

		      case 'd':	/* toggle debug flag */
			break;

		      case 'g':	/* return the value of the CPU registers */
			ptr = remcomOutBuffer;
			ptr = mem2hex((char*) &regs->std.r0, ptr, 9*8, 0);	/* v0..t7 */
			ptr = mem2hex((char*) &regs->r9, ptr, 7*8, 0);		/* s0..s6 */
			ptr = mem2hex((char*) &regs->std.r16, ptr, 3*8, 0);	/* a0..a2 */
			ptr = mem2hex((char*) &regs->std.r19, ptr, 10*8, 0);	/* a3..at */
			ptr = mem2hex((char*) &regs->std.gp, ptr, 1*8, 0);	/* gp */
			ptr = mem2hex((char*) &regs->sp, ptr, 1*8, 0);		/* sp */
			private_memset(ptr, '0', 8*2); ptr += 8*2;		/* zero */
			ptr = mem2hex((char*) regs->fpu, ptr, 32*8, 0);		/* f0..f31 */
			ptr = mem2hex((char*) &regs->std.pc, ptr, 1*8, 0);	/* pc */
			break;

		      case 'G':	/* set the value of the CPU registers - return OK */
			ptr = &remcomInBuffer[1];
			ptr = hex2mem(ptr, (char*) &regs->std.r0, 9*8, 0);	/* v0..t7 */
			ptr = hex2mem(ptr, (char*) &regs->r9, 7*8, 0);		/* s0..s6 */
			ptr = hex2mem(ptr, (char*) &regs->std.r16, 3*8, 0);	/* a0..a2 */
			ptr = hex2mem(ptr, (char*) &regs->std.r19, 10*8, 0);	/* a3..at */
			ptr = hex2mem(ptr, (char*) &regs->std.gp, 1*8, 0);	/* gp */
			ptr = hex2mem(ptr, (char*) &regs->sp, 1*8, 0);		/* sp */
			ptr += 8*2;						/* zero */
			ptr = hex2mem(ptr, (char*) regs->fpu, 32*8, 0);		/* f0..f31 */
			ptr = hex2mem(ptr, (char*) &regs->std.pc, 1*8, 0);	/* pc */
			strcpy(remcomOutBuffer, "OK");
			break;

		      case 'H':	/* set thread */
			/* not currently supported */
			sprintf(remcomOutBuffer, "E04, Set thread unsupported");
			break;

		      case 'P':	/* set the value of one CPU register - return OK */
			ptr = &remcomInBuffer[1];
			if (hex2int(&ptr, &regno) && *ptr++ == '=' && regno <= VFP) {
				reg_addr = regno2addr(regno, regs);
				DBG(DBG_STORE_REG, ("kgdb: storing %s to reg %d @ %p (regs=%p)\n",
						    ptr, regno, reg_addr, regs));
				if (reg_addr) {
					hex2mem(ptr, (char*)reg_addr, 8, 0);
				}
				strcpy(remcomOutBuffer, "OK");
			} else {
				strcpy(remcomOutBuffer, "E01");
			}
			break;

		      case 'm':	/* mAA..AA,LLLL  Read LLLL bytes at address AA..AA */
			ptr = &remcomInBuffer[1];
			if (hex2long(&ptr, &addr) && *ptr++ == ',' && hex2int(&ptr, &length)) {
				DBG(DBG_MEM, ("kgdb: reading %d bytes at %lx\n", length, addr));
				if (mem2hex((char *)addr, remcomOutBuffer, length, 1)) {
					break;
				}
				strcpy(remcomOutBuffer, "E03");
			} else {
				strcpy(remcomOutBuffer, "E01");
			}
			break;

		      case 'M':
			/* MAA..AA,LLLL: Write LLLL bytes at address AA.AA return OK */

			ptr = &remcomInBuffer[1];
			if (hex2long(&ptr, &addr) && *ptr++ == ','
			    && hex2int(&ptr, &length) && *ptr++ == ':')
			{
				DBG(DBG_MEM, ("kgdb: writing %d bytes %s at %lx\n",
					      length, ptr, addr));
				if (hex2mem(ptr, (char *)addr, length, 1)) {
					strcpy(remcomOutBuffer, "OK");
				} else {
					strcpy(remcomOutBuffer, "E03");
				}
			} else {
				strcpy(remcomOutBuffer, "E02");
			}
			break;

		      case 's':	/* sAA..AA    Step at address AA..AA (optional) */
		      case 'c':	/* cAA..AA    Continue at address AA..AA (optional) */
			/* try to read optional parameter, pc unchanged if no parm: */
			ptr = &remcomInBuffer[1];
			if (hex2long(&ptr, &addr)) {
				regs->std.pc = addr;
			} /* if */
			DBG(DBG_SINGLE_STEP, ("kgdb: continue/step (%c) at %lx\n",
					      remcomInBuffer[0], regs->std.pc));
			if (remcomInBuffer[0] == 's' && !set_breakpoint(regs)) {
				strcpy(remcomOutBuffer, "E03");
				break;
			}
			/*
			 * Need to flush the instruction cache here,
			 * as we may have deposited a breakpoint, and
			 * the icache has no way of knowing that a
			 * data ref to some location may have changed
			 * something that is in the instruction cache.
			 */
			imb();
			return;

		      case 'k':	/* kill the program */
			break;

		      case 'r':	/* reset */
			halt();
			break;

		      case 'q':	/* general query request */
			if (strcmp (&remcomInBuffer[1], "Offsets") == 0) {
#if 0
				/*
				 * This is how qOffsets is supposed to
				 * work.  However, it just doesn't
				 * work.  It maybe because OSF/1 uses
				 * ecoff (see comments in remote.c).
				 *
				 * It doesn't seem to be very useful
				 * anyways.
				 */
				extern void _ftext(), _fdata(), _fbss();

				sprintf(remcomOutBuffer,
					"Text=%lx;Data=%lx;Bss=%lx",
					(long)_ftext/2, (long)_fdata/2,
					(long)_fbss/2);
#endif
			} else {
				sprintf(remcomOutBuffer,
					"E04, Unknown query %s",
					&remcomInBuffer[1]);
			}
			break;

		      default:
			printk("kgdb: handle_exception: unknown command `%c'\n",
			       remcomInBuffer[0]);
			break;
		}

		/* reply to the request */
		putpacket(remcomOutBuffer);
	}
}


/*
 * This function will generate a breakpoint exception.  It is used at
 * the beginning of a program to sync up with a debugger and can be
 * used otherwise as a quick means to stop program execution and
 * "break" into the debugger.
 */
void kgdb_breakpoint(void)
{
	if (!initialized) {
		return;
	}
	asm volatile("call_pal %0" :: "i"(PAL_bpt));
}


void kgdb_enter (void)
{
	printk("\nkgdb: waiting for gdb to connect...\n");
	kgdb_breakpoint();
}


unsigned long kgdb_init (unsigned long mem_start, unsigned long mem_end)
{
	extern int kgdb_enabled;

	initialized = 1;
	initDebugIO();

	if (kgdb_enabled & 0x1) {
		kgdb_enter();
	}

#ifdef CONFIG_KGDB_TRACING
	/* abuse profiling buffer as a tracebuffer: */
	prof_len = 128*1024;	/* 512KB of tracing memory */
	if (mem_start + prof_len * sizeof(prof_buffer[0]) < mem_end) {
		prof_buffer = (u32*) mem_start;
		mem_start += prof_len * sizeof(prof_buffer[0]);
	} else {
		printk("kgdb_init: insufficient memory for tracing buffer\n");
	}
#endif
	return mem_start;
}


void
kgdb_init2 (void)
{
	request_irq(4 - SERIAL_LINE, irq_handler, SA_INTERRUPT, "kgdb", NULL);
}


#ifdef CONFIG_KGDB_TRACING

void
kgdb_start_tracing (u_long id)
{
	unsigned long *lbase, len;
#	define ROUND_UP(v,a)	(((v) + (a) - 1) & ~((a) - 1))
#	define ROUND_DOWN(v,a)	((v) & ~((a) - 1))

	len = (prof_len - 1) * sizeof(prof_buffer[0]);
	lbase = (unsigned long *) ROUND_UP((unsigned long) (prof_buffer + 1), sizeof(long));
	len -= (char *) lbase - (char *) (prof_buffer + 1);
	len = ROUND_DOWN(len, 4*sizeof(long));

	DBG(DBG_TRACING, ("kgdb: trace_buf=(id=%p, base=%p, limit=%p, len=%ld)\n",
			  id, lbase, lbase + len / sizeof(long), len));
	trace_buf.id   = id;
	trace_buf.base = lbase;
	trace_buf.next = lbase;
	save_flags(trace_buf.limit);
	cli();	/* don't allow interrupts when collecting trace */

	/* important to set this last: */
	trace_buf.limit = lbase + len / sizeof(long);

	/* now start tracing: */
	kgdb_breakpoint();
}


long
kgdb_stop_tracing (u_long id)
{
	if (id == trace_buf.id) {
		long len;

		trace_buf.limit = 0;	/* atomically stop tracing */

		len = trace_buf.next - trace_buf.base;
		DBG(DBG_TRACING, ("kgdb: tracing recorded %ld words\n", len));
		trace_buf.id = 0;
		restore_flags(trace_buf.flags);	/* re-set interrupt level */
		return len * sizeof(trace_buf.base[0]);
	}
	DBG(DBG_TRACING, ("kgdb_stop_tracing: got id %lx when expecting %lx\n",
			  id, trace_buf.id));
	return 0;
}

#endif /* CONFIG_KGDB_TRACING */
